/*
 * Copyright (c) 1996-1999 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */





/************************************************************************
 * utils.c: libthread functions common to all modules.
************************************************************************/





/************************************************************************
 * header files.
************************************************************************/

#include <errno.h>
#include <stdarg.h>

#include "threadP.h"





/************************************************************************
 * constants.
************************************************************************/

static const char* THR_TRACEFLAG_ENVSTR = "THR_TRACEFLAG";
static const char* THR_TRACEFILE_ENVSTR = "THR_TRACEFILE";









/************************************************************************
 * bool thr_is_ready_io(const WaitSet& ws, bool block, thread_t last_tid
 *							thread_t* tidp, tag_t* tagp)
 *
 * check to see if any of the I/O bindables specified in `ws' are
 * ready for reading. `block' specifies if the function should block
 * until an item becomes ready.
 *
 * `last_tid' represents that point that the search should
 * commence - it may be treated as a hint, since this information 
 * can't be incorporated with the current Windows NT I/O multiplexing
 * mechanism.
 *
 * '*tidp' and '*tagp' specify the ready source.
************************************************************************/

bool
thr_is_ready_io(const WaitSet& ws,
				bool block,
				thread_t last_tid,
				thread_t* tidp,
				tag_t* tagp)
{
	WaitSet::WaitReturn	ret;


    assert(thr_curr);

	
	// wait for indication of bindable ready to be read
	while( true )
	{
		ret = ws.Wait( block, last_tid, tidp, tagp );

		if( ret == WaitSet::WaitInput )
		{
			// there's ready input - *tidp and *tagp specify it
			return true;
		}
		else if( ret == WaitSet::WaitError )
		{
			// there was an error during the wait
			thr_report_error( "WaitSet::Wait failed.\n" );
			return false;
		}
		else if( ret == WaitSet::WaitTimeout )
		{
			// if we specified a block, we shouldn't timeout
			assert( !block );
			return false;
		}
		else
		{
			// NOTREACHED
			assert( false );
		}
	}

    /* NOTREACHED */
    thr_report_error("*** select inconsistency ***\n");
    abort();
    return false;
}








/************************************************************************
 * void thr_service_io(void)
 *
 * wait for a non-blocking read on one of the global file descriptors.
 * once a file becomes ready, unblock the thread bound to the ready
 * file. if the file is not-special, read the waiting message into
 * the receiver's buffer.
 *
 * if the `ready_q' is empty, then this function blocks indefinitely
 * waiting for a file/signal message. if ready threads are available,
 * a poll is done to see if a currently blocked thread can be moved onto
 * the ready queue.
************************************************************************/

void
thr_service_io(void)
{
    bool		do_block;
    bool		input_ready;
	thread_t	tid;
	tag_t		tag;
    Thread		t;


    DO_TRACE_1(("thr_service_io(): waiting for message"));

	// check for input, blocking if we don't have anything else to do
    do_block = ISEMPTYQ(thr_ready_q);
    input_ready = thr_is_ready_io(thr_all_waitset, do_block, thr_last, &tid, &tag);
	assert( input_ready || !do_block );

	// consume the input if a thread is blocked waiting for it
	if( input_ready )
	{
		if( tid != THR_TID_UNSPEC )
		{
			// find the thread that has this bindable bound to it
			assert( thr_tab->IsThread( thr_tab->GetBindable(tid)->tid ));
			t = thr_tab->GetThread( thr_tab->GetBindable(tid)->tid );
		}
		else
		{
			// the input source wasn't a bindable
			//
			// currently this should only be messages under Windows NT
			// all other asynchronous sources are represented by bindables

			// TODO - any handling here? or is this an error condition?
		}

		// release the thread if it is blocked waiting for input
		if( t->flags.blocked )
		{
			// deal with the input
			*(t->msg.tidp)	= tid;
			*(t->msg.tagp)	= tag;

			// consume the input if desired
			if( !(t->flags.polling) )
			{
				if( tid != THR_TID_UNSPEC )
				{
					// the input was a bindable - handle the message
					thr_tab->GetBindable(tid)->HandleMessage(t->msg.buf, t->msg.nitemsp);
				}
				else
				{
					// the input source wasn't a bindable - 
					// TODO - consume the Windows message into a thread msg
				}
			}

			// move the thread to the ready queue
			t->flags.blocked = false;
			t->flags.polling = false;
			ENQUEUE(thr_ready_q, t);
		}
    }
}





/************************************************************************
 * void thr_print_tcb(void)
 *
 * dump out the components of the current thread control block.
************************************************************************/

void
thr_print_tcb(void)
{
    unsigned i;

    assert(thr_curr);

    fprintf(stderr, "struct thread {\n");
    fprintf(stderr, "    id {\n");
    fprintf(stderr, "        tid         %u\n", thr_curr->id.tid);
    fprintf(stderr, "        ptid        %u\n", thr_curr->id.ptid);
    fprintf(stderr, "        errno       %d\n", thr_curr->id.errno_local);
    fprintf(stderr, "        name        %s\n", thr_curr->id.name);
    fprintf(stderr, "    }\n");
    fprintf(stderr, "    flags {\n");
    fprintf(stderr, "        is_main     %u\n", thr_curr->flags.is_main);
    fprintf(stderr, "        own_st      %u\n", thr_curr->flags.own_st);
    fprintf(stderr, "        started     %u\n", thr_curr->flags.started);
    fprintf(stderr, "        exited      %u\n", thr_curr->flags.exited);
    fprintf(stderr, "        killed      %u\n", thr_curr->flags.killed);
    fprintf(stderr, "        suspended   %u\n", thr_curr->flags.suspended);
    fprintf(stderr, "        blocked     %u\n", thr_curr->flags.blocked);
    fprintf(stderr, "        polling     %u\n", thr_curr->flags.polling);
    fprintf(stderr, "        waiting     %u\n", thr_curr->flags.waiting);
    fprintf(stderr, "        reaped      %u\n", thr_curr->flags.reaped);
    fprintf(stderr, "    }\n");
    fprintf(stderr, "    exec {\n");
    fprintf(stderr, "        state       %p\n", thr_curr->exec.state);
    fprintf(stderr, "        stack       %p\n", thr_curr->exec.stack);
    fprintf(stderr, "        st_bot      %p\n", thr_curr->exec.st_bot);
    fprintf(stderr, "        st_top      %p\n", thr_curr->exec.st_top);
    fprintf(stderr, "        stack_size  %u\n", thr_curr->exec.stack_size);
    fprintf(stderr, "    }\n");
    fprintf(stderr, "    call {\n");
    fprintf(stderr, "        func        %p\n", thr_curr->call.func);
    fprintf(stderr, "        args        %p\n", thr_curr->call.args);
    fprintf(stderr, "        result      %p\n", thr_curr->call.result);
    fprintf(stderr, "    }\n");
    fprintf(stderr, "    wait {\n");
    fprintf(stderr, "        waiter      %u\n", thr_curr->wait.waiter);
    fprintf(stderr, "        departed    %p\n", thr_curr->wait.departed);
    fprintf(stderr, "        status      %p\n", thr_curr->wait.status);
    fprintf(stderr, "    }\n");
    fprintf(stderr, "    key {\n");
    fprintf(stderr, "        key_data = { ");
    for (i = 0; i < THR_KEY_TABLE_SIZE; i++) {
        fprintf(stderr, " %p ", thr_curr->key.key_data[i]);
    }
    fprintf(stderr, "}\n    }\n");
    fprintf(stderr, "    msg {\n");
    fprintf(stderr, "        mq          %p\n", thr_curr->msg.mq);
	fprintf(stderr, "        tidp        %p\n", thr_curr->msg.tidp);
	fprintf(stderr, "        tagp        %p\n", thr_curr->msg.tagp);
    fprintf(stderr, "        buf         %p\n", thr_curr->msg.buf);
    fprintf(stderr, "        nitemsp     %p\n", thr_curr->msg.nitemsp);
    fprintf(stderr, "    }\n");
    fprintf(stderr, "    fd {\n");
    fprintf(stderr, "        waitset     %p\n", &(thr_curr->fd.waitset));
    fprintf(stderr, "        last_file   %d\n", thr_curr->fd.last_file);
    fprintf(stderr, "        last_socket %d\n", thr_curr->fd.last_socket);
    fprintf(stderr, "    sig {\n");
    fprintf(stderr, "        nsigs       %u\n", thr_curr->sig.nsigs);
    fprintf(stderr, "    }\n");
    fprintf(stderr, "}   curr            %p\n", thr_curr);

    return;
}




/************************************************************************
 * void thr_report_error(const char* fmt, ....)
 *
 * Formats and dumps the given message to the appropriate error output.
************************************************************************/

void
thr_report_error(const char* fmt, ...)
{
	char	msg_buf[PRINTF_BUFSIZE];
	va_list	args;


	/* format the output message */
    va_start(args, fmt);
    vsprintf(msg_buf, fmt, args);
    va_end(args);

	fprintf(stderr, msg_buf);
}





/************************************************************************
 * void thr_init(void)
 *
 * initializes the thread package. this function should be called
 * _BEFORE_ any other thread library function.
 *
 * queues for different kinds of threads are allocated and initial
 * setup is done here. a pseudo-thread is created for the main process.
 *
 * debugging/tracing support for the thread library is provided by
 * environment variables. the variable THR_TRACEFLAG enables tracing and
 * the variable THR_TRACEFILE specifies the destination of the trace
 * info.
************************************************************************/

void
thr_init(void)
{
    unsigned i;
    char*    env_str;

    if (thr_init_done) {
        return;
    }

	// ensure that the thread table exists
	thr_tab = new ThrTable( THR_TAB_SIZE );

    NEWQ(thr_ready_q);
    NEWQ(thr_zombie_q);
    NEWQ(thr_waiter_q);


	/* create a new thread to represent the main thread */
	thr_curr = Tcb::Alloc( 0, NULL, NULL, 0, NULL, 0, 1 );

    thr_init_arch_magic();

#if defined(i386_unknown_nt4_0)
	thr_pid = GetCurrentProcessId();
#else
    thr_pid = getpid();
#endif /* defined(i386_unknown_nt4_0) */

    if ((env_str = getenv(THR_TRACEFLAG_ENVSTR))) {
        thr_trace_flag = atoi(env_str);
        thr_trace_flag++;
    }

    thr_trace_fp = stderr;
    if ((env_str = (char *) getenv(THR_TRACEFILE_ENVSTR))) {
        if (!(thr_trace_fp = fopen(env_str, "a"))) {
            perror(env_str);
            thr_trace_fp = stderr;
        }
    }

    for (i = 0; i < THR_KEY_TABLE_SIZE; i++) {
        thr_key_destructors[i] = (Destructor) 0;
    }

    THR_ENABLE_PREEMPTION();

	// the ordering of these two statements is important - the
	// output uses the thread library, so it must be called
	// after the library thinks it has been initialized
    thr_init_done = 1;

    DO_TRACE_1(("thr_init(): thread library initialized, main tcb %p",
        thr_curr));

    return;
}




