/*
 * Copyright (c) 1996-1999 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */
#include "threadP.h"
#include "waitset.h"
#include <strings.h> // For bzero (referenced here as FD_ZERO)


bool
WaitSet::Contains( thread_t tid ) const
{
	return FD_ISSET( thr_tab->GetIOBindable(tid)->GetWaiterDesc(), &fds );
}


bool
WaitSet::Add( thread_t tid )
{
	if( !Contains(tid) )
	{
		FD_SET(thr_tab->GetIOBindable(tid)->GetWaiterDesc(), &fds);
		nWaitees++;
	}
	return true;
}


void
WaitSet::Remove( thread_t tid )
{
	if( Contains(tid) )
	{
		FD_CLR(thr_tab->GetIOBindable(tid)->GetWaiterDesc(), &fds);
		nWaitees--;
	}
}



void
WaitSet::Remove( const WaitSet& ws )
{
	WaitSetIterator	iter( ws );
	thread_t tid;

	while( iter.Next( &tid ) )
	{
		Remove( tid );
	}
}


void
WaitSet::Clear( void )
{
	FD_ZERO(&fds);
	nWaitees = 0;
}


WaitSet::WaitReturn
WaitSet::Wait( bool block,
				thread_t last_tid_hint, 
				thread_t* tidp, tag_t* tagp ) const
{
	struct		timeval	poll_timeout = {0, 0};
	fd_set		tmpfds;
	int			waitRet;
	WaitReturn	ret;
	thread_t	tid;
	bool		found;

	
	// initialize the out-parameters
	assert( tidp != NULL );
	assert( tagp != NULL );
	*tidp = THR_TID_UNSPEC;
	*tagp = MSG_TAG_UNSPEC;

	// check for one of our handles to be signalled, blocking if desired
	tmpfds = fds;
	waitRet = select( FD_SETSIZE, &tmpfds, NULL, NULL, (block ? NULL : &poll_timeout) );

	// determine why we regained control
	if( waitRet > 0 )
	{
		// there's input to be read
		ret = WaitSet::WaitInput;

		// determine which descriptor was signalled
		// TODO - take last_tid hint into account
		WaitSetIterator	iter( *this );
		found = false;
		while( iter.Next( &tid ) )
		{
			if( FD_ISSET( thr_tab->GetIOBindable( tid )->GetWaiterDesc(), &tmpfds ) )
			{
				// we found a signalled descriptor
				found = true;
				break;
			}
		}
		if( !found )
		{
			thr_report_error( "*** select inconsistency ***\n" );
			abort();
		}

		*tidp = tid;
		*tagp = thr_tab->GetBindable( tid )->GetTag();
	}
	else if( waitRet == 0 )
	{
		// the call timed out
		ret = WaitSet::WaitTimeout;
	}
	else
	{
		// the call failed
		ret = WaitSet::WaitError;
	}

	return ret;
}



bool
WaitSetIterator::Next( thread_t* tidp )
{
	// find the next item in the set
	while( (index < FD_SETSIZE) && !FD_ISSET( (int)index, &(ws.fds) ) )
	{
		index++;
	}
	bool found = (index < FD_SETSIZE);

	if( found )
	{
		// return the item's tid if desired
		if( tidp != NULL )
		{
			// we don't know the type of item represented by index,
			// so we try each option in order
			Socket* ps = Socket::Find( (PDDESC)index, tidp );
			if( ps == NULL )
			{
				File* pf = File::Find( (PDDESC)index, tidp );
				if( pf == NULL )
				{
					// we didn't find an item, even though we should've
					thr_report_error( "*** WaitSetIterator inconsistency ***\n" );
					abort();
				}
			}
		}

		// advance to next index
		index++;
	}

	return found;
}


